import inspect
from abc import ABC
from contextlib import suppress
from functools import reduce, wraps
from itertools import chain
from logging import getLogger
from pathlib import Path
from types import FunctionType
from typing import TYPE_CHECKING, Optional, Union, get_args, get_origin
from urllib.parse import urlencode

from mopidy.models import Album, Artist, Image, Playlist, Ref, SearchResult, Track
from requests import get

if TYPE_CHECKING:  # pragma: no cover
    from backend import TidalBackend

__all__ = ["login_hack", "speak_login_hack"]

NoneType = type(None)

UNION_TYPES = {Union}

try:  # pragma: no cover
    from types import UnionType

    UNION_TYPES |= {UnionType}
except ImportError:  # pragma: no cover
    pass

logger = getLogger(__name__)


def extract_types(possibly_union_type) -> list:
    """Extract the real type from an optional type."""
    if get_origin(possibly_union_type) in UNION_TYPES:
        return nonnull_types(possibly_union_type)
    return [possibly_union_type]


def nonnull_types(t):
    return [x for x in get_args(t) if x is not NoneType]


def interesting_types(t) -> set:
    """Find all the interesting types in a type, as a flat list."""
    if base_type := get_origin(t):
        if base_type is list:
            return set(chain.from_iterable(interesting_types(x) for x in get_args(t)))
        if base_type is dict:
            return interesting_types(get_args(t)[1])
    return {t}


class Builder(ABC):
    mapping: dict

    def build(self, t):
        if supertype := get_origin(t):
            subtypes = nonnull_types(t)
            if supertype is list:
                return [self.build(subtypes[0])]
            elif supertype is set:
                return {self.build(subtypes[0])}
            else:
                assert supertype is dict
                k, v = subtypes
                return {self.build(k): self.build(v)}
        else:
            return self.mapping[t]()


class ObjectBuilder(Builder):
    width = height = 150

    def __init__(self, *_, schema: str, uri: str, url: str, msg: str, **kwargs):
        self.schema = schema
        self.uri = uri
        self.url = url
        self.msg = msg
        self.mapping = {
            str: self._login_uri,
            Playlist: lambda: Playlist(
                uri="tidal:playlist:login",
                name=self.msg,
                tracks=[self.build(Track)],
            ),
            Ref: lambda: Ref(
                name=self.msg,
                type=self.ref_type(),
                uri=self.uri,
            ),
            Ref.playlist: lambda: Ref.playlist(
                name=self.msg,
                uri="tidal:playlist:login",
            ),
            Track: lambda: Track(uri="tidal:track:login", name=self.msg),
            Artist: lambda: Artist(uri="tidal:artist:login", name=self.msg),
            Album: lambda: Album(uri="tidal:album:login", name=self.msg),
            Image: lambda: Image(
                uri=self._image_url(), width=self.width, height=self.height
            ),
            SearchResult: lambda: SearchResult(
                artists=[self.build(Artist)],
                albums=[self.build(Album)],
                tracks=[self.build(Track)],
            ),
        }

    def _image_url(self) -> str:
        """Link to a qr code encoding the login url."""
        return "https://api.qrserver.com/v1/create-qr-code/?" + urlencode(
            dict(size=f"{self.width}x{self.height}", data=self.url)
        )

    def ref_type(self):
        return "directory" if self.schema.endswith("s") else self.schema

    def _login_uri(self) -> str:
        return self.uri if self.uri else f"tidal:{self.schema}:login"


class PassthroughBuilder(Builder):
    def __init__(self, *_, by_type: dict):
        self.mapping = by_type


def doublewrap(fn):
    """Double decorate to allow version with/without args.

    See https://stackoverflow.com/a/14412901/15452601
    """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        without_args = (
            len(args) == 1 and not kwargs and isinstance(args[0], FunctionType)
        )
        if without_args:
            return fn(args[0])
        else:
            # pass args/kwargs through to decorated fn
            return lambda f: fn(f, *args, **kwargs)

    return wrapper


def find_uri(args_mapping, kwargs):
    uri = None
    for k in ("uri", "uris"):
        for mapping in (args_mapping, kwargs):
            if v := mapping.get(k):
                uri = v
    if isinstance(uri, list):
        uri = uri[0]

    return uri


@doublewrap
def login_hack(fn, type=None, passthrough=False):
    manual_return_type = type

    @wraps(fn)
    def wrapper(obj, *args, **kwargs):
        backend: "TidalBackend" = obj.backend
        if not backend.logged_in and backend.login_method == "HACK":
            schema = ""
            uri = ""
            expected_return_type = NoneType
            spec = inspect.getfullargspec(fn)
            # we only need the first vararg anyhow, but this solution is not general
            spec_args = spec.args + [spec.varargs]
            args_mapping = {
                k: next(iter(args[i : i + 1]), None)
                for i, k in enumerate(spec_args[1:])
            }
            if "uri" in repr(spec):
                uri = find_uri(args_mapping, kwargs)
                if uri:
                    _, schema, *_ = uri.split(":")
            elif "field" in spec.args:
                schema = kwargs.get("field", args_mapping["field"])

            if schema:
                type_mapping = {
                    "artists": Artist,
                    "albums": Album,
                    "playlists": Playlist,
                    "tracks": Track,
                    "moods": Ref,
                    "mixes": Ref,
                    "genres": Ref,
                }
                plural = (
                    schema
                    if schema.endswith("s")
                    else schema + ("es" if schema.endswith("x") else "s")
                ).replace("my_", "")
                expected_return_type = type_mapping.get(plural, NoneType)

            if manual_return_type:
                return_type = manual_return_type
            else:
                # Assume all declared return types have the same structure
                declared_return_types = extract_types(fn.__annotations__["return"])
                possible_return_types = reduce(
                    lambda a, b: a | interesting_types(b),
                    [set(), *declared_return_types],
                )
                expected_type_possible = (
                    expected_return_type is not NoneType
                    and interesting_types(expected_return_type).issubset(
                        possible_return_types
                    )
                )
                return_type = (
                    match_structure(declared_return_types[0], expected_return_type)
                    if expected_type_possible
                    else declared_return_types[0]
                )

            url = backend.login_url
            msg = f"Please visit {url} to log in."

            logger.info("Not logged in. " + msg)
            if passthrough:
                return PassthroughBuilder(by_type={str: lambda: msg}).build(return_type)
            else:
                return ObjectBuilder(schema=schema, uri=uri, url=url, msg=msg).build(
                    return_type
                )
        elif backend.login_method == "HACK":
            audio_helper = LoginAudioHelper(backend)
            audio_helper.remove()

        return fn(obj, *args, **kwargs)

    return wrapper


def match_structure(target_type, inner_type):
    """Return a type struture equivalent to the target but with the right inner type."""
    base_type = get_origin(target_type)
    if base_type is dict:
        return base_type[get_args(target_type)[0], inner_type]
    if base_type is list:
        return base_type[inner_type]
    else:
        return inner_type


voice_rss_api_key = "eb909fe9f2ce403bb7209de172d096f1"


def speech_url(msg: str) -> str:
    return "https://api.voicerss.org?" + urlencode(
        dict(
            key=voice_rss_api_key,
            hl="en-gb",
            c="OGG",
            f="16khz_16_bit_stereo",
            src=msg,
        )
    )


class LoginAudioHelper:
    def __init__(self, backend: "TidalBackend"):
        self.backend = backend
        self.outdir = backend.data_dir / "login_audio"
        url = self.backend.login_url
        self._audiof = None
        if url:
            *_, code = url.split("/")
            self._audiof = self.outdir / f"{code}.ogg"

    def remove(self):
        if self.outdir.exists():
            for x in self.outdir.glob("*.ogg"):
                x.unlink()

    def download(self, url: str) -> Optional[str]:
        self.outdir.mkdir(parents=True, exist_ok=True)
        r = get(url)
        try:
            r.raise_for_status()
        except Exception:
            return None
        assert self._audiof
        with self._audiof.open("wb") as f:
            f.write(r.content)
        return self._audiof.as_uri()


def speak_login_hack(fn):
    @wraps(fn)
    def wrapper(obj, *args, **kwargs):
        backend: "TidalBackend" = obj.backend
        if not backend.logged_in and backend.login_method == "HACK":
            audio_helper = LoginAudioHelper(backend)
            url = backend.login_url
            msg = f"Please visit {url}, log in, and add this device.   Then come back and refresh to remove this message."
            return audio_helper.download(speech_url(msg))
        else:
            return fn(obj, *args, **kwargs)

    return wrapper



from mopidy.models import Track

watermark = " [TIDAL]"
mock_track = Track(uri="tidal:track:0:0:0", artists=[], name=None)


def apply_watermark(val):
    return val + watermark


def remove_watermark(watermarked_val):
    if watermarked_val is None:
        return None

    if watermarked_val.endswith(watermark):
        watermarked_val = watermarked_val[: -len(watermark)]

    return watermarked_val



_ctx = {
    "config": None,
}


def set_config(cfg):
    _ctx["config"] = cfg


def get_config():
    if not _ctx["config"]:
        raise ValueError("Extension configuration not set.")
    return _ctx["config"]



from __future__ import unicode_literals

from typing import TYPE_CHECKING

from login_hack import speak_login_hack

if TYPE_CHECKING:  # pragma: no cover
    from mopidy_tidal.backend import TidalBackend

import logging
from pathlib import Path

from mopidy import backend
from tidalapi import Quality
from tidalapi.media import ManifestMimeType

from . import Extension, context

logger = logging.getLogger(__name__)


class TidalPlaybackProvider(backend.PlaybackProvider):
    backend: "TidalBackend"

    @speak_login_hack
    def translate_uri(self, uri):
        logger.info("TIDAL uri: %s", uri)
        parts = uri.split(":")
        track_id = int(parts[4])
        session = self.backend.session
        if session.config.quality == Quality.hi_res_lossless:
            if "HIRES_LOSSLESS" in session.track(track_id).media_metadata_tags:
                logger.info("Playback quality: %s", session.config.quality)
            else:
                logger.info(
                    "No HIRES_LOSSLESS available for this track; Using playback quality: %s",
                    "LOSSLESS",
                )

        stream = session.track(track_id).get_stream()
        manifest = stream.get_stream_manifest()
        logger.info("MimeType:{}".format(stream.manifest_mime_type))
        logger.info(
            "Starting playback of track:{}, (quality:{}, codec:{}, {}bit/{}Hz)".format(
                track_id,
                stream.audio_quality,
                manifest.get_codecs(),
                stream.bit_depth,
                stream.sample_rate,
            )
        )

        if stream.manifest_mime_type == ManifestMimeType.MPD:
            data = stream.get_manifest_data()
            if data:
                mpd_path = Path(
                    Extension.get_cache_dir(context.get_config()), "manifest.mpd"
                )
                with open(mpd_path, "w") as file:
                    file.write(data)

                return "file://{}".format(mpd_path)
            else:
                raise AttributeError("No MPD manifest available!")
        elif stream.manifest_mime_type == ManifestMimeType.BTS:
            urls = manifest.get_urls()
            if isinstance(urls, list):
                return urls[0]
            else:
                return urls



from __future__ import unicode_literals

import logging
from concurrent.futures import ThreadPoolExecutor
from contextlib import suppress
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union

from mopidy import backend, models
from mopidy.models import Image, Ref, SearchResult, Track
from requests.exceptions import HTTPError
from tidalapi.exceptions import ObjectNotFound, TooManyRequests
from tidalapi.workers import get_items

from mopidy_tidal import full_models_mappers, ref_models_mappers
from mopidy_tidal.login_hack import login_hack
from mopidy_tidal.lru_cache import LruCache
from mopidy_tidal.playlists import PlaylistMetadataCache
from mopidy_tidal.utils import apply_watermark

if TYPE_CHECKING:  # pragma: no cover
    from mopidy_tidal.backend import TidalBackend

logger = logging.getLogger(__name__)


class ImagesGetter:
    def __init__(self, session):
        self._session = session
        self._image_cache = LruCache(directory="image")

    @staticmethod
    def _log_image_not_found(obj):
        logger.debug(
            'No images available for %s "%s"',
            type(obj).__name__,
            getattr(obj, "name", getattr(obj, "title", getattr(obj, "id"))),
        )

    @classmethod
    def _get_image_uri(cls, obj):
        method = None

        if hasattr(obj, "image"):
            if hasattr(obj, "picture") and getattr(obj, "picture", None) is not None:
                method = obj.image
            elif (
                hasattr(obj, "square_picture")
                and getattr(obj, "square_picture", None) is not None
            ):
                method = obj.image
            elif hasattr(obj, "cover") and getattr(obj, "cover", None) is not None:
                method = obj.image
            elif hasattr(obj, "images") and getattr(obj, "images", None) is not None:
                # Mix types contain images type with three small/medium/large image sizes
                method = obj.image
            else:
                # Handle artists/albums/playlists/mixes with missing images
                cls._log_image_not_found(obj)
                return
        else:
            cls._log_image_not_found(obj)
            return

        dimensions = (750, 640, 480)
        for dim in dimensions:
            args = (dim,)
            try:
                return method(*args)
            except ValueError:
                pass

        cls._log_image_not_found(obj)

    def _get_api_getter(self, item_type: str):
        return getattr(self._session, item_type, None)

    def _get_images(self, uri) -> List[Image]:
        assert uri.startswith("tidal:"), f"Invalid TIDAL URI: {uri}"

        parts = uri.split(":")
        item_type = parts[1]
        if item_type == "track":
            # For tracks, retrieve the artwork of the associated album
            item_type = "album"
            item_id = parts[3]
            uri = ":".join([parts[0], "album", parts[3]])
        elif item_type == "album":
            item_id = parts[2]
        elif item_type == "playlist":
            item_id = parts[2]
        elif item_type == "artist":
            item_id = parts[2]
        elif item_type == "mix":
            item_id = parts[2]
        else:
            # uri has no image associated to it (eg. tidal:mood tidal:genres etc.)
            return []

        if uri in self._image_cache:
            # Cache hit
            logger.debug("Cache hit for {}".format(uri))
            return self._image_cache[uri]

        logger.debug("Retrieving %r from the API", uri)
        getter = self._get_api_getter(item_type)
        if not getter:
            logger.warning("The API item type %s has no session getters", item_type)
            return []

        item = getter(item_id)
        if not item:
            logger.debug("%r is not available on the backend", uri)
            return []

        img_uri = self._get_image_uri(item)
        if not img_uri:
            logger.debug("%r has no associated images", uri)
            return []

        logger.debug("Image URL for %r: %r", uri, img_uri)
        return [Image(uri=img_uri, width=320, height=320)]

    def __call__(self, uri: str) -> Tuple[str, List[Image]]:
        parts = uri.split(":")
        item_type = parts[1]
        if item_type not in ["artist", "album", "playlist", "mix", "track"]:
            logger.debug("URI %s type has no image getters", uri)
            return uri, []
        try:
            return uri, self._get_images(uri)
        except (AssertionError, AttributeError, ObjectNotFound) as err:
            logger.error("%s when processing URI %r: %s", type(err), uri, err)
            return uri, []
        except (HTTPError, TooManyRequests) as err:
            logger.error("%s when processing URI %r: %s", type(err), uri, err)
            return uri, []

    def cache_update(self, images):
        self._image_cache.update(images)


class TidalLibraryProvider(backend.LibraryProvider):
    root_directory = models.Ref.directory(uri="tidal:directory", name="Tidal")
    backend: "TidalBackend"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._artist_cache = LruCache()
        self._album_cache = LruCache()
        self._track_cache = LruCache()
        self._playlist_cache = PlaylistMetadataCache()

    @property
    def _session(self):
        return self.backend.session

    @staticmethod
    def _convert_tracks(
        tracks: Union[
            dict, list[dict], list[Track], list[tuple[str, Union[Any, None]]]
        ],
    ) -> list[Track]:
        """
        Convert looked up tracks to a list of Track objects.

        This is to ensure compatibility between Mopidy < 4.0 (which expects a
        list of Track objects) and Mopidy >= 4.0 (which expects a list of key-value
        tuples or a dictionary).
        """
        if isinstance(tracks, list) and all(isinstance(t, Track) for t in tracks):
            return tracks  # type: ignore[return-value]
        if isinstance(tracks, dict):
            # Single-track lookup returns a dictionary
            tracks = [tracks]  # type: ignore[assignment]
        if isinstance(tracks, list) and all(isinstance(t, tuple) for t in tracks):
            # Single-track lookup returns a list of tuples that need to be converted
            # to a dictionary first
            tracks = [dict(tracks)]  # type: ignore[assignment]
        if isinstance(tracks, list) and all(isinstance(t, dict) for t in tracks):
            return [
                Track(
                    **{
                        k: t.get(k)  # type: ignore[assignment]
                        for k in Track.model_fields  # pylint: disable=no-member
                        if k != "model"
                    },
                )
                for t in tracks
            ]

        raise TypeError(
            f"Unsupported track format: {type(tracks)}. "
            "Expected list of Track objects, list of tuples, or a dictionary."
        )

    @login_hack(passthrough=True)
    def get_distinct(self, field, query=None) -> set[str]:
        from mopidy_tidal.search import tidal_search

        logger.debug("Browsing distinct %s with query %r", field, query)
        session = self._session

        if not query:  # library root
            if field in {"artist", "albumartist"}:
                return {
                    apply_watermark(a.name) for a in session.user.favorites.artists()
                }
            elif field == "album":
                return {
                    apply_watermark(a.name) for a in session.user.favorites.albums()
                }
            elif field in {"track", "track_name"}:
                return {
                    apply_watermark(t.name) for t in session.user.favorites.tracks()
                }
        else:
            if field == "artist":
                return {
                    apply_watermark(a.name) for a in session.user.favorites.artists()
                }
            elif field in {"album", "albumartist"}:
                artists, _, _ = tidal_search(session, query=query, exact=True)
                if len(artists) > 0:
                    artist = artists[0]
                    artist_id = artist.uri.split(":")[2]
                    return {
                        apply_watermark(a.name)
                        for a in self._get_artist_albums(session, artist_id)
                    }
            elif field in {"track", "track_name"}:
                return {
                    apply_watermark(t.name) for t in session.user.favorites.tracks()
                }
            pass

        return set()

    @login_hack
    def browse(self, uri) -> list[Ref]:
        logger.info("Browsing uri %s", uri)
        if not uri or not uri.startswith("tidal:"):
            return []

        session = self._session

        # summaries

        if uri == self.root_directory.uri:
            return ref_models_mappers.create_root()

        elif uri == "tidal:my_artists":
            return ref_models_mappers.create_artists(
                session.user.favorites.artists_paginated()
            )
        elif uri == "tidal:my_albums":
            return ref_models_mappers.create_albums(
                session.user.favorites.albums_paginated()
            )
        elif uri == "tidal:my_playlists":
            return self.backend.playlists.as_list()
        elif uri == "tidal:my_mixes":
            return ref_models_mappers.create_mixes(session.user.favorites.mixes())
        elif uri == "tidal:my_tracks":
            return ref_models_mappers.create_tracks(
                session.user.favorites.tracks_paginated()
            )
        elif uri == "tidal:home":
            return ref_models_mappers.create_category_directories(
                uri, session.home(use_legacy_endpoint=True)
            )
        elif uri == "tidal:for_you":
            return ref_models_mappers.create_category_directories(
                uri, session.for_you()
            )
        elif uri == "tidal:explore":
            return ref_models_mappers.create_category_directories(
                uri, session.explore()
            )
        elif uri == "tidal:hires":
            return ref_models_mappers.create_category_directories(
                uri, session.hires_page()
            )
        elif uri == "tidal:moods":
            return ref_models_mappers.create_moods(session.moods())
        elif uri == "tidal:mixes":
            return ref_models_mappers.create_mixes([m for m in session.mixes()])
        elif uri == "tidal:genres":
            return ref_models_mappers.create_genres(session.genre.get_genres())

        # Category nested on a page (eg. page(For You).category[0..n])
        # These have 3-part uris
        with suppress(ValueError):
            _, page_id, type, category_id = uri.split(":")
            category = session.page.get(f"pages/{page_id}").categories[int(category_id)]
            return ref_models_mappers.create_mixed_directory(category.items)

        # details with 2-part uris
        try:
            _, type, id = uri.split(":")

            if type == "album":
                return ref_models_mappers.create_tracks(
                    self._get_album_tracks(session, id)
                )

            elif type == "artist":
                top_10_tracks = ref_models_mappers.create_tracks(
                    self._get_artist_top_tracks(session, id)[:10]
                )

                albums = ref_models_mappers.create_albums(
                    self._get_artist_albums(session, id)
                )

                return albums + top_10_tracks

            elif type == "playlist":
                return ref_models_mappers.create_tracks(
                    self._get_playlist_tracks(session, id)
                )

            elif type == "mood":
                return ref_models_mappers.create_playlists(
                    self._get_mood_items(session, id)
                )

            elif type == "genre":
                return ref_models_mappers.create_playlists(
                    self._get_genre_items(session, id)
                )

            elif type == "mix":
                return ref_models_mappers.create_tracks(
                    self._get_mix_tracks(session, id)
                )

            elif type == "page":
                return ref_models_mappers.create_mixed_directory(session.page.get(id))
            else:
                return []

        except ValueError:
            logger.exception("Unable to parse uri '%s' for browse.", uri)
            return []
        except HTTPError:
            logger.exception("Unable to retrieve object from uri '%s'", uri)
            return []

    @login_hack
    def search(self, query=None, uris=None, exact=False) -> Optional[SearchResult]:
        from mopidy_tidal.search import tidal_search

        try:
            artists, albums, tracks = tidal_search(
                self._session, query=query, exact=exact
            )
            return SearchResult(artists=artists, albums=albums, tracks=tracks)
        except Exception as ex:
            logger.info("EX")
            logger.info("%r", ex)

    @login_hack
    def get_images(self, uris) -> dict[str, list[Image]]:
        logger.info("Searching Tidal for images for %r" % uris)
        images_getter = ImagesGetter(self._session)

        with ThreadPoolExecutor(4, thread_name_prefix="mopidy-tidal-images-") as pool:
            pool_res = pool.map(images_getter, uris)

        images = {uri: item_images for uri, item_images in pool_res if item_images}
        images_getter.cache_update(images)
        return images

    @login_hack
    def lookup(self, uris=None) -> list[Track]:
        if isinstance(uris, str):
            uris = [uris]
        if not hasattr(uris, "__iter__"):
            uris = [uris]

        tracks = []
        cache_updates = {}

        for uri in uris or []:
            data = []
            try:
                parts = uri.split(":")
                item_type = parts[1]
                cache_name = f"_{parts[1]}_cache"
                cache_miss = True

                try:
                    data = getattr(self, cache_name)[uri]
                    cache_miss = not bool(data)
                except (AttributeError, KeyError):
                    pass

                if cache_miss:
                    try:
                        lookup = getattr(self, f"_lookup_{parts[1]}")
                    except AttributeError:
                        continue

                    data = cache_data = lookup(self._session, parts)
                    cache_updates[cache_name] = cache_updates.get(cache_name, {})
                    if item_type == "playlist":
                        # Playlists should be persisted on the cache as objects,
                        # not as lists of tracks. Therefore, _lookup_playlist
                        # returns a tuple that we need to unpack
                        data, cache_data = data

                    cache_updates[cache_name][uri] = cache_data

                if item_type == "playlist" and not cache_miss:
                    tracks += data.tracks
                else:
                    tracks += data if hasattr(data, "__iter__") else [data]
            except HTTPError as err:
                logger.error("%s when processing URI %r: %s", type(err), uri, err)

        for cache_name, new_data in cache_updates.items():
            getattr(self, cache_name).update(new_data)

        tracks = self._convert_tracks(tracks)
        self._track_cache.update({track.uri: track for track in tracks})
        logger.info("Returning %d tracks", len(tracks))
        return tracks

    @classmethod
    def _get_playlist_tracks(cls, session, playlist_id):
        try:
            pl = session.playlist(playlist_id)
        except ObjectNotFound:
            logger.debug("No such playlist: %s", playlist_id)
            return []
        return pl.tracks_paginated()

    @staticmethod
    def _get_genre_items(session, genre_id):
        from tidalapi.playlist import Playlist

        filtered_genres = [g for g in session.genre.get_genres() if genre_id == g.path]
        if filtered_genres:
            return filtered_genres[0].items(Playlist)
        return []

    @staticmethod
    def _get_mood_items(session, mood_id):
        filtered_moods = [
            m for m in session.moods() if mood_id == m.api_path.split("/")[-1]
        ]

        if filtered_moods:
            mood = filtered_moods[0].get()
            return [p for p in mood]
        return []

    @staticmethod
    def _get_mix_tracks(session, mix_id):
        try:
            mix = session.mix(mix_id)
        except ObjectNotFound:
            logger.debug("No such mix: %s", mix_id)
            return []
        return mix.items()

    def _lookup_playlist(self, session, parts):
        playlist_id = parts[2]
        tidal_playlist = session.playlist(playlist_id)
        tidal_tracks = self._get_playlist_tracks(session, playlist_id)
        pl_tracks = full_models_mappers.create_mopidy_tracks(tidal_tracks)
        pl = full_models_mappers.create_mopidy_playlist(tidal_playlist, pl_tracks)
        # We need both the list of tracks and the mapped playlist object for
        # caching purposes
        return pl_tracks, pl

    @staticmethod
    def _get_artist_albums(session, artist_id):
        try:
            artist = session.artist(artist_id)
        except ObjectNotFound:
            logger.debug("No such artist: %s", artist_id)
            return []
        return artist.get_albums()

    @staticmethod
    def _get_album_tracks(session, album_id):
        try:
            album = session.album(album_id)
        except ObjectNotFound:
            logger.debug("No such album: %s", album_id)
            return []
        return album.tracks()

    def _lookup_track(self, session, parts):
        if len(parts) == 3:  # Track in format `tidal:track:<track_id>`
            track_id = parts[2]
            track = session.track(track_id)
            album_id = str(track.album.id)
        else:  # Track in format `tidal:track:<artist_id>:<album_id>:<track_id>`
            album_id = parts[3]
            track_id = parts[4]

        try:
            tracks = self._get_album_tracks(session, album_id)
        except ObjectNotFound:
            logger.warning("No such album: %s", album_id)
            tracks = []
        except TooManyRequests:
            logger.warning("Too many requests when fetching album: %s", album_id)
            tracks = []

        # If album is unavailable, no tracks will be returned
        if tracks:
            track = next((t for t in tracks if t.id == int(track_id)), None)
            if not track:
                # A StopIteration is rare here, but it can happen - e.g. if the
                # cache is stale and the track has been removed from Tidal
                logger.warning("No such track: %s", track_id)  # pragma: no cover
                return []  # Return early to prevent accessing attributes of None

            return [full_models_mappers.create_mopidy_track(None, None, track)]
        else:
            return []

    def _lookup_album(self, session, parts):
        album_id = parts[2]
        tracks = self._get_album_tracks(session, album_id)

        return full_models_mappers.create_mopidy_tracks(tracks)

    @staticmethod
    def _get_artist_top_tracks(session, artist_id):
        return session.artist(artist_id).get_top_tracks()

    def _lookup_artist(self, session, parts):
        artist_id = parts[2]
        tracks = self._get_artist_top_tracks(session, artist_id)
        return full_models_mappers.create_mopidy_tracks(tracks)



from __future__ import unicode_literals

import logging

from mopidy.models import Album, Artist, Playlist, Track

from mopidy_tidal.helpers import to_timestamp

logger = logging.getLogger(__name__)


def _get_release_date(obj):
    d = None
    for attr in ("release_date", "tidal_release_date"):
        d = getattr(obj, attr, None)
        if d:
            break

    if d:
        return str(d.year)


def create_mopidy_artists(tidal_artists):
    return [create_mopidy_artist(a) for a in tidal_artists]


def create_mopidy_artist(tidal_artist):
    if tidal_artist is None:
        return None

    return Artist(uri="tidal:artist:" + str(tidal_artist.id), name=tidal_artist.name)


def create_mopidy_albums(tidal_albums):
    return [create_mopidy_album(a, None) for a in tidal_albums]


def create_mopidy_album(tidal_album, artists):
    if artists:
        pass
    elif tidal_album.artists:
        artists = [create_mopidy_artist(artist) for artist in tidal_album.artists]
    else:
        artists = [create_mopidy_artist(tidal_album.artist)]

    return Album(
        uri="tidal:album:" + str(tidal_album.id),
        name=tidal_album.name,
        artists=artists,
        date=_get_release_date(tidal_album),
    )


def create_mopidy_tracks(tidal_tracks):
    return [create_mopidy_track(None, None, t) for t in tidal_tracks]


def create_mopidy_track(artist, album, tidal_track):
    uri = "tidal:track:{0}:{1}:{2}".format(
        tidal_track.artist.id, tidal_track.album.id, tidal_track.id
    )
    if artist is None:
        if tidal_track.artists:
            artist = [create_mopidy_artist(artist) for artist in tidal_track.artists]
        else:
            artist = [create_mopidy_artist(tidal_track.artist)]
    if album is None:
        album = create_mopidy_album(tidal_track.album, None)

    track_len = tidal_track.duration * 1000
    return Track(
        uri=uri,
        name=tidal_track.full_name,
        track_no=tidal_track.track_num,
        artists=artist,
        album=album,
        length=track_len,
        date=_get_release_date(tidal_track),
        # Different attribute name for disc_num on tidalapi >= 0.7.0
        disc_no=getattr(tidal_track, "disc_num", getattr(tidal_track, "volume_num")),
    )


def create_mopidy_playlist(tidal_playlist, tidal_tracks):
    return Playlist(
        uri=f"tidal:playlist:{tidal_playlist.id}",
        name=tidal_playlist.name,
        tracks=tidal_tracks,
        last_modified=to_timestamp(tidal_playlist.last_updated),
    )


def create_mopidy_mix_playlist(tidal_mix):
    return Playlist(
        uri=f"tidal:mix:{tidal_mix.id}",
        name=f"{tidal_mix.title} ({tidal_mix.sub_title})",
        tracks=create_mopidy_tracks(tidal_mix.items()),
    )



import threading
from functools import partial
from http.server import BaseHTTPRequestHandler, HTTPServer
from string import whitespace
from typing import Callable, Optional
from urllib.parse import unquote

HTML_BODY = """<!DOCTYPE html>
<html>
<head>
<title>TIDAL Web Auth</title>
</head>
<body>

<h1>KEEP THIS TAB OPEN</h1>
<a href={authurl} target="_blank" rel="noopener noreferrer">Click here to be forwarded to TIDAL Login page</a>
{interactive}

</body>
</html>
""".format

INTERACTIVE_HTML_BODY = """
<p>...then, after login, copy the complete URL of the page you were redirected to.</p>
<p>Probably a "Oops / Not found" page, nevertheless we need the whole URL as is.</p>
<form method="post">
  <label for="code">Paste the response URL here:</label>
  <input type="url" id="code" name="code">
  <input type="submit" value="Submit">
</form>
"""


class WebAuthServer:
    def __init__(self):
        self.handler: Optional[partial] = None
        self.callback: Optional[Callable] = None
        self.response_code: str = ""
        self.daemon_started: bool = False

    def start_oauth_daemon(self, login_url: str, port: int, pkce_enabled: bool):
        if self.daemon_started:
            return

        self.handler = partial(
            HTTPHandler, login_url, self.set_response_code, pkce_enabled
        )

        daemon = threading.Thread(
            name="TidalOAuthLogin",
            target=HTTPServer(("", port), self.handler).serve_forever,
        )
        daemon.daemon = (
            True  # Set as a daemon so it will be killed once the main thread is dead.
        )
        daemon.start()
        self.daemon_started = True

    def set_callback(self, callback: Callable[[str], None]):
        self.callback = callback

    def set_response_code(self, response_code: str):
        self.response_code = response_code
        if self.callback:
            self.callback(response_code)

    @property
    def is_daemon_running(self):
        return self.daemon_started

    @property
    def get_response_code(self):
        if self.response_code == "":
            return None
        else:
            return self.response_code


class HTTPHandler(BaseHTTPRequestHandler, object):
    def __init__(
        self,
        login_url: str,
        callback: Callable[[str], None],
        pkce_enabled: bool,
        *args,
        **kwargs
    ):
        self.login_url = login_url
        self.callback_fn: Callable = callback
        self.pkce_enabled = pkce_enabled
        super().__init__(*args, **kwargs)

    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        interactive = INTERACTIVE_HTML_BODY if self.pkce_enabled else ""
        self.wfile.write(
            HTML_BODY(authurl=self.login_url, interactive=interactive).encode()
        )

    def do_POST(self):
        content_length = int(self.headers.get("Content-Length"), 0)
        body = self.rfile.read(content_length).decode()
        try:
            form = {k: v for k, v in (p.split("=", 1) for p in body.split("&"))}
            code_url = unquote(form["code"].strip(whitespace))
        except:
            self.send_response(400)
            self.end_headers()
            self.wfile.write(b"Malformed request")
            raise
        else:
            self.callback_fn(code_url)
            self.send_response(200)
            self.end_headers()
            self.wfile.write(b"Success!\nEnjoy your music!")



from __future__ import unicode_literals

import logging
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from enum import IntEnum
from typing import (
    Callable,
    Collection,
    Iterable,
    List,
    Mapping,
    Sequence,
    Tuple,
    Type,
    Union,
)

from lru_cache import SearchCache
from tidalapi.album import Album
from tidalapi.artist import Artist
from tidalapi.media import Track

from mopidy_tidal.full_models_mappers import (
    create_mopidy_albums,
    create_mopidy_artists,
    create_mopidy_tracks,
)
from mopidy_tidal.utils import remove_watermark

logger = logging.getLogger(__name__)


class SearchField(IntEnum):
    ANY = 0
    ARTIST = 1
    ALBUMARTIST = 2
    ALBUM = 3
    TITLE = 4


@dataclass
class SearchFieldMeta:
    field: SearchField
    request_field: str
    results_fields: Sequence[str]
    model_classes: Collection[Type[Union[Artist, Album, Track]]]
    mappers: Sequence[Callable[[Collection], Sequence]]


fields_meta = {
    meta.field: meta
    for meta in [
        SearchFieldMeta(
            SearchField.ANY,
            request_field="any",
            results_fields=("artists", "albums", "tracks"),
            model_classes=(Artist, Album, Track),
            mappers=(create_mopidy_artists, create_mopidy_albums, create_mopidy_tracks),
        ),
        SearchFieldMeta(
            SearchField.ARTIST,
            request_field="artist",
            results_fields=("artists",),
            model_classes=(Artist,),
            mappers=(create_mopidy_artists,),
        ),
        SearchFieldMeta(
            SearchField.ALBUMARTIST,
            request_field="albumartist",
            results_fields=("artists",),
            model_classes=(Artist,),
            mappers=(create_mopidy_artists,),
        ),
        SearchFieldMeta(
            SearchField.ALBUM,
            request_field="album",
            results_fields=("albums",),
            model_classes=(Album,),
            mappers=(create_mopidy_albums,),
        ),
        SearchFieldMeta(
            SearchField.TITLE,
            request_field="track_name",
            results_fields=("tracks",),
            model_classes=(Track,),
            mappers=(create_mopidy_tracks,),
        ),
    ]
}


def _get_flattened_query_and_field_meta(
    query: Mapping[str, str]
) -> Tuple[str, SearchFieldMeta]:
    q = " ".join(
        query[field]
        for field in ("any", "artist", "album", "track_name")
        if query.get(field)
    )

    fields_by_request_field = {
        field_meta.request_field: field_meta for field_meta in fields_meta.values()
    }

    matched_field_meta = fields_by_request_field["any"]
    for attr in ("track_name", "album", "artist", "albumartist"):
        field_meta = fields_by_request_field.get(attr)
        if field_meta and query.get(attr):
            matched_field_meta = field_meta
            break

    return q, matched_field_meta


def _get_exact_result(
    query: Mapping,
    results: Tuple[Iterable[Artist], Iterable[Album], Iterable[Track]],
    field_meta: SearchFieldMeta,
) -> Tuple[List[Artist], List[Album], List[Track]]:
    query_value = query[field_meta.request_field]
    filtered_results = [], [], []

    for i, attr in enumerate(
        (SearchField.TITLE, SearchField.ALBUM, SearchField.ARTIST)
    ):
        if attr == field_meta.field:
            item = next(
                (
                    res
                    # TODO: why not results[-i-1]?
                    for res in results[len(results) - i - 1]
                    if res.name and res.name.lower() == query_value.lower()
                ),
                None,
            )

            if item:
                filtered_results[len(results) - i - 1].append(item)
            break

    return filtered_results


def _expand_artist_top_tracks(artist: Artist) -> List[Track]:
    return artist.get_top_tracks(limit=25)


def _expand_album_tracks(album: Album) -> List[Track]:
    return album.tracks()


def _expand_results_tracks(
    results: Tuple[List[Artist], List[Album], List[Track]],
) -> Tuple[List[Artist], List[Album], List[Track]]:
    results_ = list(results)
    artists = results_[0]
    albums = results_[1]

    with ThreadPoolExecutor(4, thread_name_prefix="mopidy-tidal-search-") as pool:
        pool_res = pool.map(_expand_artist_top_tracks, artists)
        for tracks in pool_res:
            results_[2].extend(tracks)

        pool_res = pool.map(_expand_album_tracks, albums)
        for tracks in pool_res:
            results_[2].extend(tracks)

    # Remove any duplicate tracks from results
    tracks_by_id = OrderedDict({track.id: track for track in results_[2]})
    results_[2] = list(tracks_by_id.values())
    return tuple(results_)


@SearchCache
def tidal_search(session, query, exact=False):
    logger.info("Searching Tidal for: %r", query)
    query = query.copy()

    for field, value in query.items():
        if hasattr(value, "__iter__") and not isinstance(value, (str, bytes)):
            value = value[0]
        query[field] = remove_watermark(value)

    query_string, field_meta = _get_flattened_query_and_field_meta(query)

    results = [[], [], []]  # artists, albums, tracks
    api_results = session.search(query_string, models=field_meta.model_classes)

    for i, field_type in enumerate(
        (SearchField.ARTIST, SearchField.ALBUM, SearchField.TITLE)
    ):
        meta = fields_meta[field_type]
        results_field = meta.results_fields[0]
        mapper = meta.mappers[0]

        if not (results_field in api_results and results_field in meta.results_fields):
            continue

        results[i] = api_results[results_field]

    if exact:
        results = list(_get_exact_result(query, tuple(results), field_meta))

    _expand_results_tracks(results)
    for i, field_type in enumerate(
        (SearchField.ARTIST, SearchField.ALBUM, SearchField.TITLE)
    ):
        meta = fields_meta[field_type]
        mapper = meta.mappers[0]
        results[i] = mapper(results[i])

    return tuple(results)



from __future__ import unicode_literals

import logging
import time
from concurrent.futures import Future
from pathlib import Path
from typing import Optional, Union

from mopidy import backend
from pykka import ThreadingActor
from tidalapi import Config, Session
from tidalapi import __version__ as tidalapi_ver

from mopidy_tidal import Extension
from mopidy_tidal import __version__ as mopidy_tidal_ver
from mopidy_tidal import context, library, playback, playlists
from mopidy_tidal.web_auth_server import WebAuthServer

logger = logging.getLogger(__name__)


class TidalBackend(ThreadingActor, backend.Backend):
    def __init__(self, config, audio):
        super().__init__()
        # Mopidy cfg
        self._config = config
        context.set_config(self._config)
        self._tidal_config = config[Extension.ext_name]

        # Backend
        self.playback = playback.TidalPlaybackProvider(audio=audio, backend=self)
        self.library = library.TidalLibraryProvider(backend=self)
        self.playlists = playlists.TidalPlaylistsProvider(backend=self)

        # Session parameters
        self._active_session: Optional[Session] = None
        self._logged_in: bool = False
        self.uri_schemes: tuple[str] = ("tidal",)
        self._login_future: Optional[Future] = None
        self._login_url: Optional[str] = None
        self.data_dir: Path = Path(Extension.get_data_dir(self._config))
        self.session_file_path: Path = Path("")
        self.web_auth_server: WebAuthServer = WebAuthServer()

        # Config parameters
        # Lazy: Connect lazily, i.e. login only when user starts browsing TIDAL directories
        self.lazy_connect: bool = False
        # Login Method:
        #   BLOCK:       Immediately prompt user for login (This will block mopidy startup!)
        #   HACK/AUTO:   Display dummy track with login URL. When clicked, QR code and TTS is generated
        self.login_method: str = "BLOCK"
        # pkce_enabled: If true, TIDAL session will use PKCE auth. Otherwise OAuth2 is used
        self.auth_method: str = "OAUTH"
        self.pkce_enabled: bool = False
        # login_server_port: Port to use for login HTTP server, eg. <host_ip>:<port>. Default <host_ip>:8989
        self.login_server_port: int = 8989

    @property
    def session(self):
        if not self.logged_in:
            self._login()
        return self._active_session

    @property
    def logged_in(self):
        if not self._logged_in:
            if self._active_session.load_session_from_file(self.session_file_path):
                logger.info("Loaded TIDAL session from file %s", self.session_file_path)
                self._logged_in = self.session_valid
        return self._logged_in

    @property
    def session_valid(self):
        # Returns true when session is logged in and valid
        return self._active_session.check_login()

    def on_start(self):
        logger.info("Mopidy-Tidal version: v%s", mopidy_tidal_ver)
        logger.info("Python-Tidal version: v%s", tidalapi_ver)
        quality = self._tidal_config["quality"]
        client_id = self._tidal_config["client_id"]
        client_secret = self._tidal_config["client_secret"]
        self.auth_method = self._tidal_config["auth_method"]
        if self.auth_method == "PKCE":
            self.pkce_enabled = True
        self.login_server_port = self._tidal_config["login_server_port"]
        logger.info("PKCE login web server port: %s", self.login_server_port)
        self.login_method = self._tidal_config["login_method"]
        if self.login_method == "AUTO":
            # Add AUTO as alias to HACK login method
            self.login_method = "HACK"
        self.lazy_connect = self._tidal_config["lazy"]
        logger.info("Quality: %s", quality)
        logger.info("Authentication: %s", "PKCE" if self.pkce_enabled else "OAuth")
        config = Config(quality=quality)

        # Set the session filename, depending on the type of session
        if self.pkce_enabled:
            self.session_file_path = Path(self.data_dir, "tidal-pkce.json")
        else:
            self.session_file_path = Path(self.data_dir, "tidal-oauth.json")

        if (self.login_method == "HACK") and not self._tidal_config["lazy"]:
            logger.warning("AUTO login implies lazy connection, setting lazy=True.")
            self.lazy_connect = True
        logger.info(
            "Login method: %s", "BLOCK" if self.pkce_enabled == "BLOCK" else "AUTO"
        )

        if client_id and client_secret:
            logger.info("Using client id & client secret from config")
            config.client_id = client_id
            config.client_secret = client_secret
        elif (client_id and not client_secret) or (client_secret and not client_id):
            logger.warning("Always provide both client_id and client_secret")
            logger.info("Using default client id & client secret from python-tidal")
        else:
            logger.info("Using default client id & client secret from python-tidal")

        self._active_session = Session(config)
        if not self.lazy_connect:
            self._login()

    def _login(self):
        """Load session at startup or create a new session"""
        if self._active_session.load_session_from_file(self.session_file_path):
            logger.info(
                "Loaded existing TIDAL session from file %s...", self.session_file_path
            )
        if not self.session_valid:
            if not self.login_server_port:
                # A. Default login, user must find login URL in Mopidy log
                logger.info("Creating new session (OAuth)...")
                self._active_session.login_oauth_simple(fn_print=logger.info)
            else:
                # B. Interactive login, user must perform login using web auth
                logger.info(
                    "Creating new session (%s)...",
                    "PKCE" if self.pkce_enabled else "OAuth",
                )
                if self.pkce_enabled:
                    # PKCE Login
                    login_url = self._active_session.pkce_login_url()
                    logger.info(
                        "Please visit 'http://localhost:%s' to authenticate",
                        self.login_server_port,
                    )
                    # Enable web server for interactive login + callback on form Submit
                    self.web_auth_server.set_callback(self._web_auth_callback)
                    self.web_auth_server.start_oauth_daemon(
                        login_url, self.login_server_port, self.pkce_enabled
                    )
                else:
                    # OAuth login
                    login_url = self.login_url
                    logger.info(
                        "Please visit 'http://localhost:%s' or '%s' to authenticate",
                        self.login_server_port,
                        login_url,
                    )
                    # Enable web server for interactive login (no callback)
                    self.web_auth_server.start_oauth_daemon(
                        login_url, self.login_server_port, self.pkce_enabled
                    )

                # Wait for user to complete interactive login sequence
                max_time = time.time() + 300
                while time.time() < max_time:
                    if self._logged_in:
                        if not self.pkce_enabled:
                            self._complete_login()
                        return
                    logger.info(
                        "Time left to complete authentication: %s sec",
                        int(max_time - time.time()),
                    )
                    time.sleep(5)
                raise TimeoutError("You took too long to log in")

    def _web_auth_callback(self, url_redirect: str):
        """Callback triggered on web auth completion
        :param url_redirect: URL of the 'Ooops' page, where the user was redirected to after login.
        :type url_redirect: str
        """
        if self.pkce_enabled:
            try:
                # Query for auth tokens
                json: dict[str, Union[str, int]] = (
                    self._active_session.pkce_get_auth_token(url_redirect)
                )
                # Parse and set tokens.
                self._active_session.process_auth_token(json, is_pkce_token=True)
                self._logged_in = True
            except:
                raise ValueError("Response code is required for PKCE login!")
        # Store session after auth completion
        self._complete_login()

    def _complete_login(self):
        """Perform final steps of login sequence; save session to file"""
        if self.session_valid:
            # Only store current session if valid
            logger.info("TIDAL Login OK")
            self._active_session.save_session_to_file(self.session_file_path)
            self._logged_in = True
        else:
            logger.error("TIDAL Login Failed")
            raise ConnectionError("Failed to log in.")

    @property
    def logging_in(self) -> bool:
        """Are we currently waiting for user confirmation to log in?"""
        return bool(self._login_future and self._login_future.running())

    @property
    def login_url(self) -> Optional[str]:
        """Start a new login sequence (if not active) and get the latest login URL"""
        if not self.pkce_enabled:
            if not self._logged_in and not self.logging_in:
                login_url, self._login_future = self._active_session.login_oauth()
                self._login_future.add_done_callback(lambda *_: self._complete_login())
                self._login_url = login_url.verification_uri_complete
            return f"https://{self._login_url}" if self._login_url else None
        else:
            if not self._logged_in and not self.web_auth_server.is_daemon_running:
                login_url = self._active_session.pkce_login_url()
                self._login_url = "http://localhost:{}".format(self.login_server_port)
                # Enable web server for interactive login + callback on form Submit
                self.web_auth_server.set_callback(self._web_auth_callback)
                self.web_auth_server.start_oauth_daemon(
                    login_url, self.login_server_port, self.pkce_enabled
                )
            return f"{self._login_url}" if self._login_url else None



from __future__ import unicode_literals

import difflib
import logging
import operator
from pathlib import Path
from threading import Event, Timer
from typing import TYPE_CHECKING, Collection, List, Optional, Tuple, Union

from mopidy import backend
from mopidy.models import Playlist as MopidyPlaylist
from mopidy.models import Ref
from requests import HTTPError
from tidalapi.playlist import Playlist as TidalPlaylist
from tidalapi.workers import get_items

from mopidy_tidal import full_models_mappers
from mopidy_tidal.full_models_mappers import create_mopidy_playlist
from mopidy_tidal.helpers import to_timestamp
from mopidy_tidal.login_hack import login_hack
from mopidy_tidal.lru_cache import LruCache
from mopidy_tidal.utils import mock_track

if TYPE_CHECKING:  # pragma: no cover
    from mopidy_tidal.backend import TidalBackend

logger = logging.getLogger(__name__)


class PlaylistCache(LruCache):
    def __getitem__(
        self, key: Union[str, TidalPlaylist], *args, **kwargs
    ) -> MopidyPlaylist:
        uri = key.id if isinstance(key, TidalPlaylist) else key
        assert uri
        uri = f"tidal:playlist:{uri}" if not uri.startswith("tidal:playlist:") else uri

        playlist = super().__getitem__(uri, *args, **kwargs)
        if (
            playlist
            and isinstance(key, TidalPlaylist)
            and to_timestamp(key.last_updated) > to_timestamp(playlist.last_modified)
        ):
            # The playlist has been updated since last time:
            # we should refresh the associated cache entry
            logger.info('The playlist "%s" has been updated: refresh forced', key.name)

            raise KeyError(uri)

        return playlist


class PlaylistMetadataCache(PlaylistCache):
    def cache_file(self, key: str) -> Path:
        return super().cache_file(key, Path("playlist_metadata"))


class TidalPlaylistsProvider(backend.PlaylistsProvider):
    backend: "TidalBackend"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._playlists_metadata = PlaylistMetadataCache()
        self._playlists = PlaylistCache()
        self._current_tidal_playlists = []
        self._playlists_loaded_event = Event()

    def _calculate_added_and_removed_playlist_ids(
        self,
    ) -> Tuple[Collection[str], Collection[str]]:
        logger.info("Calculating playlist updates..")
        session = self.backend.session

        updated_playlists = session.user.favorites.playlists_paginated()

        self._current_tidal_playlists = updated_playlists
        updated_ids = set(pl.id for pl in updated_playlists)

        if not self._playlists_metadata:
            return updated_ids, set()

        current_ids = set(uri.split(":")[-1] for uri in self._playlists_metadata.keys())
        added_ids = updated_ids.difference(current_ids)
        removed_ids = current_ids.difference(updated_ids)

        self._playlists_metadata.prune(
            *[
                uri
                for uri in self._playlists_metadata.keys()
                if uri.split(":")[-1] in removed_ids
            ]
        )

        return added_ids, removed_ids

    def _has_changes(self, playlist: MopidyPlaylist):
        upstream_playlist = self.backend.session.playlist(playlist.uri.split(":")[-1])
        if not upstream_playlist:
            return True

        upstream_last_updated_at = to_timestamp(
            getattr(upstream_playlist, "last_updated", None)
        )
        local_last_updated_at = to_timestamp(playlist.last_modified)

        if not upstream_last_updated_at:
            logger.warning(
                "You are using a version of python-tidal that does not "
                "support last_updated on playlist objects"
            )
            return True

        if upstream_last_updated_at > local_last_updated_at:
            logger.info(
                'The playlist "%s" has been updated: refresh forced', playlist.name
            )
            return True

        return False

    @login_hack(list[Ref.playlist])
    def as_list(self) -> list[Ref]:
        if not self._playlists_loaded_event.is_set():
            added_ids, _ = self._calculate_added_and_removed_playlist_ids()
            if added_ids:
                self.refresh(include_items=False)

        logger.debug("Listing TIDAL playlists..")
        refs = [
            Ref.playlist(uri=pl.uri, name=pl.name)
            for pl in self._playlists_metadata.values()
        ]

        return sorted(refs, key=operator.attrgetter("name"))

    def _lookup_mix(self, uri):
        mix_id = uri.split(":")[-1]
        session = self.backend.session
        return session.mix(mix_id)

    def _get_or_refresh_playlist(self, uri) -> Optional[MopidyPlaylist]:
        parts = uri.split(":")
        if parts[1] == "mix":
            mix = self._lookup_mix(uri)
            return full_models_mappers.create_mopidy_mix_playlist(mix)

        playlist = self._playlists.get(uri)
        if (playlist is None) or (playlist and self._has_changes(playlist)):
            self.refresh(uri, include_items=True)
        return self._playlists.get(uri)

    def create(self, name):
        tidal_playlist = self.backend.session.user.create_playlist(name, "")
        pl = create_mopidy_playlist(tidal_playlist, [])

        self._current_tidal_playlists.append(tidal_playlist)
        self.refresh(pl.uri)
        return pl

    def delete(self, uri):
        playlist_id = uri.split(":")[-1]
        session = self.backend.session

        try:
            session.request.request(
                "DELETE",
                "playlists/{playlist_id}".format(
                    playlist_id=playlist_id,
                ),
            )
        except HTTPError as e:
            # If we got a 401, it's likely that the user is following
            # this playlist but they don't have permissions for removing
            # it. If that's the case, remove the playlist from the
            # favourites instead of deleting it.
            if e.response.status_code == 401 and uri in {
                f"tidal:playlist:{pl.id}"
                for pl in session.user.favorites.playlists_paginated()
            }:
                session.user.favorites.remove_playlist(playlist_id)
            else:
                raise e

        self._playlists_metadata.prune(uri)
        self._playlists.prune(uri)

    @login_hack
    def lookup(self, uri) -> Optional[MopidyPlaylist]:
        return self._get_or_refresh_playlist(uri)

    @login_hack
    def refresh(self, *uris, include_items: bool = True) -> dict[str, MopidyPlaylist]:
        if uris:
            logger.info("Looking up playlists: %r", uris)
        else:
            logger.info("Refreshing TIDAL playlists..")

        session = self.backend.session
        plists = self._current_tidal_playlists
        mapped_playlists = {}
        playlist_cache = self._playlists if include_items else self._playlists_metadata

        for pl in plists:
            uri = "tidal:playlist:" + pl.id
            # Skip or cache hit case
            if (uris and uri not in uris) or pl in playlist_cache:
                continue

            # Cache miss case
            if include_items:
                pl_tracks = self._retrieve_api_tracks(pl)
                tracks = full_models_mappers.create_mopidy_tracks(pl_tracks)
            else:
                # Create as many mock tracks as the number of items in the playlist.
                # Playlist metadata is concerned only with the number of tracks, not
                # the actual list.
                tracks = [mock_track] * pl.num_tracks

            mapped_playlists[uri] = MopidyPlaylist(
                uri=uri,
                name=pl.name,
                tracks=tracks,
                last_modified=to_timestamp(pl.last_updated),
            )

        # When we trigger a playlists_loaded event the backend may call as_list
        # again. Set an event in playlist_cache_refresh_secs seconds to ensure
        # that we don't perform another playlist sync.
        self._playlists_loaded_event.set()
        playlist_cache_refresh_secs = self.backend._config["tidal"].get(
            "playlist_cache_refresh_secs"
        )

        if playlist_cache_refresh_secs:
            Timer(
                playlist_cache_refresh_secs,
                lambda: self._playlists_loaded_event.clear(),
            ).start()

        # Update the right playlist cache and send the playlists_loaded event.
        playlist_cache.update(mapped_playlists)
        backend.BackendListener.send("playlists_loaded")
        logger.info("TIDAL playlists refreshed")

    @login_hack
    def get_items(self, uri) -> Optional[List[Ref]]:
        playlist = self._get_or_refresh_playlist(uri)
        if not playlist:
            return

        return [Ref.track(uri=t.uri, name=t.name) for t in playlist.tracks]

    def _retrieve_api_tracks(self, playlist):
        return playlist.tracks_paginated()

    def save(self, playlist):
        old_playlist = self._get_or_refresh_playlist(playlist.uri)
        session = self.backend.session
        playlist_id = playlist.uri.split(":")[-1]
        assert old_playlist, f"No such playlist: {playlist.uri}"
        assert session, "No active session"
        upstream_playlist = session.playlist(playlist_id)

        # Playlist rename case
        if old_playlist.name != playlist.name:
            upstream_playlist.edit(title=playlist.name)

        additions = []
        removals = []
        remove_offset = 0
        diff_lines = difflib.ndiff(
            [t.uri for t in old_playlist.tracks], [t.uri for t in playlist.tracks]
        )

        for diff_line in diff_lines:
            if diff_line.startswith("+ "):
                additions.append(diff_line[2:].split(":")[-1])
            else:
                if diff_line.startswith("- "):
                    removals.append(remove_offset)
                remove_offset += 1

        # Process removals in descending order so we don't have to recalculate
        # the offsets while we remove tracks
        if removals:
            logger.info(
                'Removing %d tracks from the playlist "%s"',
                len(removals),
                playlist.name,
            )

            removals.reverse()
            for idx in removals:
                upstream_playlist.remove_by_index(idx)

        # tidalapi currently only supports appending tracks to the end of the
        # playlist
        if additions:
            logger.info(
                'Adding %d tracks to the playlist "%s"', len(additions), playlist.name
            )

            upstream_playlist.add(additions)

        # remove all defunct tracks from cache
        self._calculate_added_and_removed_playlist_ids()
        # force update the whole playlist so all state is good
        self.refresh(playlist.uri)



from __future__ import unicode_literals

import logging
import pickle
from collections import OrderedDict
from pathlib import Path
from typing import Optional

from mopidy_tidal import Extension, context

logger = logging.getLogger(__name__)


def id_to_cachef(id: str) -> Path:
    return Path(id.replace(":", "-") + ".cache")


class LruCache(OrderedDict):
    def __init__(self, max_size: Optional[int] = 1024, persist=True, directory=""):
        """
        :param max_size: Max size of the cache in memory. Set 0 or None for no
            limit (default: 1024)
        :param persist: Whether the cache should be persisted to disk
            (default: True)
        :param directory: If `persist=True`, store the cached entries in this
            subfolder of the cache directory (default: '')
        """
        super().__init__(self)
        if max_size:
            assert max_size > 0, f"Invalid cache size: {max_size}"

        self._max_size = max_size or 0
        self._cache_dir = Path(Extension.get_cache_dir(context.get_config()), directory)
        self._persist = persist
        if persist:
            self._cache_dir.mkdir(parents=True, exist_ok=True)

        self._check_limit()

    @property
    def max_size(self):
        return self._max_size

    @property
    def persist(self):
        return self._persist

    def cache_file(self, key: str, cache_dir: Optional[Path] = None) -> Path:
        parts = key.split(":")
        assert len(parts) > 2, f"Invalid TIDAL ID: {key}"
        _, obj_type, id, *_ = parts
        if not cache_dir:
            cache_dir = Path(obj_type)

        cache_dir = self._cache_dir / cache_dir / id[:2]
        cache_dir.mkdir(parents=True, exist_ok=True)
        cache_file = cache_dir / id_to_cachef(key)
        legacy_cache_file = cache_dir / f"{key}.cache"
        if legacy_cache_file.is_file():
            return legacy_cache_file

        return cache_file

    def _get_from_storage(self, key):
        cache_file = self.cache_file(key)
        err = KeyError(key)
        if not cache_file.is_file():
            # Cache miss on the filesystem
            raise err

        # Cache hit on the filesystem
        with open(cache_file, "rb") as f:
            try:
                value = pickle.load(f)
            except Exception as e:
                # If the cache entry on the filesystem is corrupt, reset it
                logger.warning(
                    "Could not deserialize cache file %s: " "refreshing the entry: %s",
                    cache_file,
                    e,
                )
                self._reset_stored_entry(key)
                raise err

        # Store the filesystem item in memory
        if value is not None:
            self.__setitem__(key, value, _sync_to_fs=False)
        logger.debug(f"Filesystem cache hit for {key}")
        return value

    def __getitem__(self, key, *_, **__):
        try:
            # Cache hit in memory
            return super().__getitem__(key)
        except KeyError as e:
            if not self.persist:
                # No persisted storage -> cache miss
                raise e

        # Check on the persisted cache
        return self._get_from_storage(key)

    def __setitem__(self, key, value, _sync_to_fs=True, *_, **__):
        if super().__contains__(key):
            del self[key]

        super().__setitem__(key, value)
        if self.persist and _sync_to_fs:
            cache_file = self.cache_file(key)
            with open(cache_file, "wb") as f:
                pickle.dump(value, f)

        self._check_limit()

    def __contains__(self, key):
        return self.get(key) is not None

    def _reset_stored_entry(self, key):
        cache_file = self.cache_file(key)
        if cache_file.is_file():
            cache_file.unlink()

    def get(self, key, default=None, *args, **kwargs):
        try:
            return self.__getitem__(key, *args, **kwargs)
        except KeyError:
            return default

    def prune(self, *keys):
        """
        Delete the specified keys both from memory and disk.
        """
        for key in keys:
            logger.debug("Pruning key %r from cache %s", key, self.__class__.__name__)

            self._reset_stored_entry(key)
            self.pop(key, None)

    def prune_all(self):
        """
        Prune all the keys in the cache.
        """
        self.prune(*[*self.keys()])

    def update(self, *args, **kwargs):
        super().update(*args, **kwargs)
        self._check_limit()

    def _check_limit(self):
        if self.max_size:
            # delete oldest entries
            while len(self) > self.max_size:
                self.popitem(last=False)


class SearchCache(LruCache):
    def __init__(self, search_function):
        super().__init__(persist=False)
        self._search_function = search_function

    def __call__(self, *args, **kwargs):
        key = str(SearchKey(**kwargs))
        cached_result = self.get(key)
        logger.info(
            "Search cache miss" if cached_result is None else "Search cache hit"
        )
        if cached_result is None:
            cached_result = self._search_function(*args, **kwargs)
            self[key] = cached_result

        return cached_result


class SearchKey(object):
    def __init__(self, **kwargs):
        fixed_query = self.fix_query(kwargs["query"])
        self._query = tuple(sorted(fixed_query.items()))
        self._exact = kwargs["exact"]
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self._exact)
            self._hash ^= hash(repr(self._query))

        return self._hash

    def __str__(self):
        return f"tidal:search:{self.__hash__()}"

    def __eq__(self, other):
        if not isinstance(other, SearchKey):
            return False

        return self._exact == other._exact and self._query == other._query

    @staticmethod
    def fix_query(query):
        """
        Removes some query parameters that otherwise will lead to a cache miss.
        Eg: 'track_no' since we can't query TIDAL for a specific album's track.
        :param query: query dictionary
        :return: sanitized query dictionary
        """
        query.pop("track_no", None)
        return query



from __future__ import unicode_literals

import logging

from mopidy.models import Ref
from tidalapi import Album, Artist, Mix, Playlist, Track
from tidalapi.mix import MixType

logger = logging.getLogger(__name__)


def create_root():
    return [
        Ref.directory(uri="tidal:home", name="Home"),
        Ref.directory(uri="tidal:for_you", name="For You"),
        Ref.directory(uri="tidal:explore", name="Explore"),
        Ref.directory(uri="tidal:hires", name="HiRes"),
        Ref.directory(uri="tidal:genres", name="Genres"),
        Ref.directory(uri="tidal:moods", name="Moods"),
        Ref.directory(uri="tidal:mixes", name="My Mixes"),
        Ref.directory(uri="tidal:my_artists", name="My Artists"),
        Ref.directory(uri="tidal:my_albums", name="My Albums"),
        Ref.directory(uri="tidal:my_playlists", name="My Playlists"),
        Ref.directory(uri="tidal:my_tracks", name="My Tracks"),
        Ref.directory(uri="tidal:my_mixes", name="Mixes & Radio"),
    ]


def create_artists(tidal_artists):
    return [create_artist(a) for a in tidal_artists]


def create_artist(tidal_artist):
    return Ref.artist(
        uri="tidal:artist:" + str(tidal_artist.id), name=tidal_artist.name
    )


def create_playlists(tidal_playlists):
    return [create_playlist(p) for p in tidal_playlists]


def create_playlist(tidal_playlist):
    return Ref.playlist(
        uri="tidal:playlist:" + str(tidal_playlist.id), name=tidal_playlist.name
    )


def create_moods(tidal_moods):
    return [create_mood(m) for m in tidal_moods]


def create_mood(tidal_mood):
    mood_id = tidal_mood.api_path.split("/")[-1]
    return Ref.directory(uri="tidal:mood:" + mood_id, name=tidal_mood.title)


def create_genres(tidal_genres):
    return [create_genre(m) for m in tidal_genres]


def create_genre(tidal_genre):
    genre_id = tidal_genre.path
    return Ref.directory(uri="tidal:genre:" + genre_id, name=tidal_genre.name)


def create_category_directories(uri, tidal_page):
    res = []
    for idx, category in enumerate(tidal_page.categories):
        if category.title == "":
            res.extend(
                [create_mixed_entry(item) for idx, item in enumerate(category.items)]
            )
        else:
            res.append(create_category_directory(uri, idx, category.title))

    # Remove None/Unsupported entries
    res_filtered = [i for i in res if i is not None]
    return res_filtered


def create_category_directory(uri, idx, name):
    return Ref.directory(uri="{}:category:{}".format(uri, idx), name=name)


def create_mixed_directory(tidal_mixed):
    res = [create_mixed_entry(m) for m in tidal_mixed]
    # Remove None/Unsupported entries
    res_filtered = [i for i in res if i is not None]
    return res_filtered


def create_mixed_entry(tidal_mixed):
    if isinstance(tidal_mixed, Mix):
        return Ref.playlist(
            uri="tidal:mix:" + tidal_mixed.id,
            name=f"{tidal_mixed.title} ({tidal_mixed.sub_title})",
        )
    elif isinstance(tidal_mixed, Album):
        return Ref.album(
            uri="tidal:album:" + str(tidal_mixed.id),
            name=f"{tidal_mixed.name} ({tidal_mixed.artist.name})",
        )
    elif isinstance(tidal_mixed, Playlist):
        return Ref.playlist(
            uri="tidal:playlist:" + str(tidal_mixed.id),
            name=f"{tidal_mixed.name}",
        )
    elif isinstance(tidal_mixed, Track):
        return create_track(tidal_mixed)
    elif isinstance(tidal_mixed, Artist):
        return create_artist(tidal_mixed)
    else:
        if hasattr(tidal_mixed, "api_path"):
            # Objects containing api_path are usually pages and must be processed further
            return Ref.directory(
                uri="tidal:page:" + tidal_mixed.api_path, name=tidal_mixed.title
            )
        elif hasattr(tidal_mixed, "artifact_id"):
            # Objects containing artifact_id can be viewed directly
            explore_id = tidal_mixed.artifact_id
            name = f"{tidal_mixed.short_header} ({tidal_mixed.short_sub_header})"
            if tidal_mixed.type == "PLAYLIST":
                return Ref.playlist(
                    uri="tidal:playlist:" + explore_id,
                    name=name,
                )
            else:
                # Unsupported type (eg. interview, exturl)
                return None
        else:
            # Unsupported type (eg. Video)
            return None


def create_mixes(tidal_mixes):
    res = [create_mix(m) for m in tidal_mixes]
    # Remove None/Unsupported entries
    res_filtered = [i for i in res if i is not None]
    return res_filtered


def create_mix(tidal_mix):
    if tidal_mix.mix_type is MixType.video_daily:
        # Skip video mixes (not supported)
        return None
    else:
        return Ref.playlist(
            uri="tidal:mix:" + tidal_mix.id,
            name=f"{tidal_mix.title} ({tidal_mix.sub_title})",
        )


def create_albums(tidal_albums):
    return [create_album(a) for a in tidal_albums]


def create_album(tidal_album):
    return Ref.album(uri="tidal:album:" + str(tidal_album.id), name=tidal_album.name)


def create_tracks(tidal_tracks):
    return [create_track(t) for t in tidal_tracks]


def create_track(tidal_track):
    uri = "tidal:track:{0}:{1}:{2}".format(
        tidal_track.artist.id, tidal_track.album.id, tidal_track.id
    )
    return Ref.track(uri=uri, name=tidal_track.name)



from __future__ import unicode_literals

import logging
import os
import sys

from mopidy import config, ext

__version__ = "0.3.12"

# TODO: If you need to log, use loggers named after the current Python module
logger = logging.getLogger(__name__)

file_dir = os.path.dirname(__file__)
sys.path.append(file_dir)


class Extension(ext.Extension):
    dist_name = "Mopidy-Tidal"
    ext_name = "tidal"
    version = __version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), "ext.conf")
        return config.read(conf_file)

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["quality"] = config.String(
            choices=["HI_RES_LOSSLESS", "LOSSLESS", "HIGH", "LOW"]
        )
        schema["client_id"] = config.String(optional=True)
        schema["client_secret"] = config.String(optional=True)
        schema["playlist_cache_refresh_secs"] = config.Integer(optional=True)
        schema["lazy"] = config.Boolean(optional=True)
        schema["login_method"] = config.String(choices=["BLOCK", "HACK", "AUTO"])
        schema["auth_method"] = config.String(optional=True, choices=["OAUTH", "PKCE"])
        schema["login_server_port"] = config.Integer(
            optional=True, choices=range(8000, 9000)
        )
        return schema

    def setup(self, registry):
        from .backend import TidalBackend

        registry.add("backend", TidalBackend)



import datetime


def to_timestamp(dt):
    if not dt:
        return 0
    if isinstance(dt, str):
        dt = datetime.datetime.fromisoformat(dt)
    if isinstance(dt, datetime.datetime):
        dt = dt.timestamp()
    return int(dt)
